-- {{{

open import Agda.Builtin.Equality
open import Agda.Builtin.Sigma
open import Agda.Primitive using (Level; lzero; lsuc; _⊔_) renaming (Set to Type)
open import Data.Nat
open import Data.Product
open import Data.Product.Base
open import Data.Product.Properties
open import Data.Bool

open import Data.Product using (_,_)
open import Function.Base using (_∘_)
open import Relation.Binary.Core
open import Relation.Binary.Definitions
open import Relation.Nullary.Negation.Core using (¬_)

import Relation.Binary.PropositionalEquality as Eq
open Eq using (_≡_; refl; trans; sym; cong; cong-app; subst)
open Eq.≡-Reasoning using (begin_; _≡⟨⟩_; step-≡; _∎)

-- Path
-- data _≡_ {l : Level} {A : Type l} : A → A → Type l where
--   refl : (x : A) → x ≡ x
path-ind : ∀ {ℓ : Level} {A : Type}
  -- Motive
  → (C : (x y : A) → x ≡ y → Type ℓ)
  -- What happens in the case of refl
  → (c : (x : A) → C x x refl)
  -- Actual path to eliminate
  → (x y : A) → (p : x ≡ y)
  -- Result
  → C x y p
path-ind C c x x refl = c x

-- Id
id : {ℓ : Level} {A : Set ℓ} → A → A
id x = x

-- Transport
transport : {X : Set} (P : X → Set) {x y : X}
  → x ≡ y
  → P x → P y
transport P refl = id

transportconst : {A : Type} {x y : A}
  → (B : Set)
  → (p : x ≡ y)
  → (b : B)
  → transport (λ _ → B) p b ≡ b
transportconst {A} {x} {y} B p b =
  let
    P : (x : A) → Type
    P = λ _ → B

    D : (x y : A) → (p : x ≡ y) → Type
    D x y p = transport P p b ≡ b

    d : (x : A) → D x x refl
    d x = refl
  in path-ind D d x y p

ap : {A B : Type} → (f : A → B) → {x y : A} → (p : x ≡ y) → f x ≡ f y
ap f {x} {y} p = path-ind (λ a b a≡b → f a ≡ f b) (λ _ → refl) x y p

-- apd
-- Lemma 2.3.4 of HoTT book
apd : {A : Type} {B : A → Type}
  -- The function that we want to apply
  → (f : (a : A) → B a)
  -- The path to apply it over
  → {x y : A} → (p : x ≡ y)
  -- Result
  → (transport B p) (f x) ≡ f y
apd {A} {B} f {x} {y} p = 
  let
    D : (x y : A) → (p : x ≡ y) → Type
    D x y p = (transport B p) (f x) ≡ f y

    d : (x : A) → D x x refl
    d x = refl
  in
  path-ind D d x y p

-- Fibers
fiber : {A B : Type} → (f : A → B) → (y : B) → Type
fiber {A} {B} f y = Σ[ x ∈ A ] f x ≡ y

-- Contractibility
isContr : Type → Type
isContr A = Σ[ x ∈ A ] (∀ y → x ≡ y)

-- Equivalence
record isEquiv {A B : Type} (f : A → B) : Type where
  field
    equiv-proof : (y : B) → isContr (fiber f y)
open isEquiv

_≃_ : (A B : Type) → Type
A ≃ B = Σ[ f ∈ (A → B) ] (isEquiv f)

-- }}}

-------------------------------------------------------------------------------
-- Exercises

lemma-212 : {A : Type} → Type
lemma-212 {A} = (x y z : A) → (x ≡ y) → (y ≡ z) → (x ≡ z)

-- Exercise 2.1: Show that the obvious proofs of Lemma 2.1.2 are pairwise equal

l212-obv-proof-1 : {A : Type} → lemma-212
l212-obv-proof-1 {A} x y z p q =
  let
    f = path-ind
      (λ a b a≡b → (z : A) → (r : b ≡ z) → a ≡ z)
      (λ a → (λ x r → r))
    r = f x y p
  in
  r z q

l212-obv-proof-2 : {A : Type} → lemma-212
l212-obv-proof-2 {A} x y z p q =
  let
    f = path-ind
      (λ a b a≡b → (x : A) → (r : x ≡ a) → x ≡ b)
      (λ a → (λ x r → r))
    r = f y z q
  in
  r x p

l212-pairwise-1-2 : {A : Type} → (x y z : A) → (x ≡ y) → (y ≡ z)
  → l212-obv-proof-1 ≡ l212-obv-proof-2
l212-pairwise-1-2 {A} x y z p q =
  let
    left-proof = l212-obv-proof-1 x y z p q
    right-proof = l212-obv-proof-2 x y z p q
  in ?

-- Exercise 2.3: Give a fourth, different proof of Lemma 2.1.2, and prove it
-- equal to the others

l212-obv-proof-4 : {A : Type} → lemma-212
l212-obv-proof-4 {A} x y z p q =
  let
    f = path-ind
      (λ a b a≡b → (x : A) → (r : x ≡ a) → x ≡ b)
      (λ a → (λ x r → r))
    r = f y z q
  in
  r x p

-- Exercise 2.4: Define, by induction on n, a general notion of n-dimensional
-- path in a type A, simultaneously with the type of boundaries of such paths

-- p : (2 ≡ 1 + 1)
-- p = refl
--
-- q : (2 ≡ 1 + 1)
-- q = ?
--
-- p2 : p ≡ q
-- q2 : p ≡ q
--
-- p3 : p2 ≡ q2
--
-- nSuc- npath 5

-- data nPath {A : Type} : ℕ → Type where
--   zeroPath : (x y : A) → (p : x ≡ y) → nPath zero
--   sucPath : {n : ℕ}
--     → (left : nPath n)
--     → (right : nPath n)
--     → (p : left ≡ right)
--     → nPath (suc n)

-- data nPath : (A : Type) → (x y : A) → Type where
--   path-zero : (A : Type) → (x y : A) → (p : x ≡ y) → nPath A x y
--   path-suc : (A : Type) → (x y : A)
--     → (left : nPath A x y)
--     → (right : nPath A x y)
--     → (p : left ≡ right)
--     → nPath (nPath A x y) left right

data nPath : Type where

-- p : nPath ℕ (1 + 1) 2
-- p = path-zero ℕ (1 + 1) 2 refl
-- 
-- q : nPath ℕ (1 + 1) 2
-- q = path-zero ℕ (1 + 1) 2 refl
-- 
-- p2 : nPath (nPath ℕ (1 + 1) 2) p q

-- Exercise 2.5: Prove that the functions (2.3.6) and (2.3.7) are
-- inverse-equivalences


-- TODO: In this implementation, should r be used in constructing the response?
-- it feels like i'm doing it wrong because r is not used
f236 : {A B : Type} {x y : A} {p : x ≡ y} {f : A → B}
  → (r : f x ≡ f y) → (transport (λ _ → B) p (f x)) ≡ f y
f236 {p = p} {f = f} _ = apd f p

f237 : {A B : Type} {x y : A} {p : x ≡ y} {f : A → B}
  → (r : (transport (λ _ → B) p (f x)) ≡ f y) → (f x ≡ f y)
f237 {p = p} {f = f} _ = ap f p

-- Exercise 2.6: Prove that if p : x = y, then the function (p ∙ –) : (y = z) →
-- (x = z) is an equivalence.

-- Exercise 2.9. Prove that coproducts have the expected universal property,
-- (A + B → X) ≃ (A → X) × (B → X).

-- Exercise 2.13
-- Show that (2 ≃ 2) ≃ 2

-- The fiber of id for the Bool y (aka all x's such that id x ≡ y, which is just
-- y)
id-fiber-base : (y : Bool) → Σ[ x ∈ Bool ] id x ≡ y
proj₁ (id-fiber-base y) = y
proj₂ (id-fiber-base y) = refl

id-fiber-is-contr : (y : Bool) → (y₁ : Σ[ x ∈ Bool ] id x ≡ y) → id-fiber-base y ≡ y₁
id-fiber-is-contr y y₁ =
  let
    y₁-Bool-equiv = snd y₁

    pair₁ : y ≡ fst y₁
    pair₁ = sym y₁-Bool-equiv

    snd-fiber : id (fst (id-fiber-base y)) ≡ y
    snd-fiber = snd (id-fiber-base y)

    pair₂ : subst (λ x → id x ≡ y) pair₁ (snd (id-fiber-base y)) ≡ snd y₁
    pair₂ =
      begin
        subst (λ x → id x ≡ y) pair₁ snd-fiber
      ≡⟨ ? ⟩
        -- subst : Substitutive {A = A} _≡_ ℓ
        -- subst P refl p = p
        -- snd-fiber
        ?
      ≡⟨ ? ⟩
        snd y₁
      ∎

    pair = pair₁ , pair₂
  in Σ-≡,≡→≡ pair

x : Substitutive ? ?
x = ?

-- Σ-≡,≡←≡ : p₁ ≡ p₂ → Σ (a₁ ≡ a₂) (λ p → subst B p b₁ ≡ b₂)
-- Σ-≡,≡←≡ refl = refl , refl

-- A proof that for every y, the fiber of id for that y (which is just going to
-- be y itself) is contractible, which means that for every pair of (a, path),
-- there needs to be a path between that pair and the pair i gave above in id-fiber-base
id-is-contr : (y : Bool) → isContr (fiber id y)
proj₁ (id-is-contr y) = id-fiber-base y
proj₂ (id-is-contr y) y₁ = id-fiber-is-contr y y₁

id-is-eqv : isEquiv id
id-is-eqv .equiv-proof y = ?

id-eqv : Bool ≃ Bool
proj₁ id-eqv = id
proj₂ id-eqv = id-is-eqv

-- proj₁ (proj₁ (proj₂ id-eqv .equiv-proof y)) = y
-- proj₂ (proj₁ (proj₂ id-eqv .equiv-proof y)) = refl
-- proj₂ (proj₂ id-eqv .equiv-proof y) y₁ = {! !}

e213 : (Bool ≃ Bool) ≃ Bool

e213-helper-1 : (y : Bool) → Σ (Bool ≃ Bool) (λ x → proj₁ e213 x ≡ y)
proj₁ (e213-helper-1 y) = {! !}
proj₂ (e213-helper-1 y) = {! !}

proj₁ e213 bool-equiv = (proj₁ bool-equiv) true
proj₁ (proj₂ e213 .equiv-proof y) = e213-helper-1 y
proj₂ (proj₂ e213 .equiv-proof y) y₁ = {! !}












 
